##
## @file evalFnlQ.R
## @brief Evaluates the marginal mean outcome of functional
##        Q-learning with a linear mean function.  Also
##        computes the marginal mean outcome of linear Q-learning
##        with ad hoc summaries.  
## @author Anonymized
## @note Anonymized
##
source ('fnlQLearn.R');
##
## @fn evalFnlPar (genData, sampleSize, M=1000, tSize=10000,
##                 nCore=8)
## @brief Evaluates the estimated value of: fnlQLearning for fnl
##        data across M training datasets of size sampleSize.
##        Evaluation is done using a test set of size tSize.
##        Note that this function works with both sparse and
##        densely observed data.  However, data must be
##        regularly spaced.  
## @param genData a fn from positive integers to list with
##        X, W, A, Y where X is the design matrix for static
##        predictors WITH AN INTERCEPT, W is the matrix of
##        fnl predictors on an equally spaced grid, A is a
##        vector of binary txts coded {-1,1}, and Y is a cts
##        outcome coded s.t. higher values are more desirable.
## @param sampleSize training set size
## @param M number of Monte Carlo replications
## @param tSize test set size
## @param nCore number of cores to use in MC iterations
## @return a list with the following key-->value pairs:
##         'fnlQVal'-->vector of length M with value attained
##         'lastQVal'-->vector of length M with value attained for
##                      std Q-learning with the last value of W used
##         'meanQVal'-->vector of length M with the value attained for
##                      std Q-learning with the mean value of W used
##
evalFnlQDense = function (genData, sampleSize, M=10000, tSize=10000,
    nCore=8){
    require ('parallel'); # old multicore
    require ('refund');
    testSet = genData (tSize);
    runSingle = function (m){
        trainSet = genData (sampleSize);
        fnlQFitted = denseQLearn (trainSet$X, trainSet$W, trainSet$A,
            trainSet$Y);
        piHatFnl = fnlQFitted$piHat (testSet$X, testSet$W);
        fnlQVal = mean (testSet$Y*(piHatFnl*testSet$A > 0));
        fnlQVal = fnlQVal/mean(piHatFnl*testSet$A > 0);
        
        stdQFitted = stdQLearn (trainSet$X, trainSet$W, trainSet$A,
            trainSet$Y);
        piHatLast = stdQFitted$piHatLast (testSet$X, testSet$W);
        lastQVal = mean (testSet$Y*(piHatLast*testSet$A > 0));
        lastQVal = lastQVal/mean(piHatLast*testSet$A > 0);
        
        piHatMean = stdQFitted$piHatMean (testSet$X, testSet$W);
        meanQVal = mean(testSet$Y*(piHatMean*testSet$A > 0));
        meanQVal = meanQVal/mean(piHatMean*testSet$A > 0);
        cat(m, " : ", fnlQVal, " , ", lastQVal, ", ",
            meanQVal, "\n");

        return (as.numeric (c (fnlQVal, lastQVal, meanQVal)));
    }
    op = mclapply (1:M, runSingle, mc.cores=nCore, mc.preschedule=T);
    fnlQVal = numeric (M);
    lastQVal = numeric (M);
    meanQVal = numeric (M);
    for (m in 1:M){
        fnlQVal[m] = op[[m]][1];
        lastQVal[m] = op[[m]][2];
        meanQVal[m] = op[[m]][3];
    }
    return (list ('fnlQVal'=fnlQVal, 'lastQVal'=lastQVal,
                  'meanQVal'=meanQVal));
}
