##
## @file fnlQLearnFixed.R
## @brief Contains fns for estimating optimal treatment regimes
##        via Q-learning with a single fnl predictor
## @author Anonymized
## @note Anonymized
##

##
## @fn denseQLearnFixed (X, W, A, Y, fz, tSeq, M=1000)
## @brief Implements fnl Q-learning at a single stage
## @param X n x p matrix of predictors (includes intercept)
## @param W n x nObs matrix of observed fnl values
## @param A n vector of txts in {-1,1}
## @param Y cts outcome, coded s.t. higher is better
## @param fz a function(al) that maps vectors z, and t to
##        the real line
## @param tSeq sequence of points upon which to sample Z(t)|W(T)
## @param M number of samples to draw to approximate E(f(Z)|W(T))
## @return a list with the following components:
##         alphaHat-->est coeff vector for main effect of X
##         nuHat-->est coeff vector linear probe of W
##         gammaHat-->est coeff vector for interaction of X and A
##         etaHat-->est coeff vector for interaction of lin probe of W and A
##         piHat-->function from X, W to a recommended txt 
##
denseQLearnFixed = function (X, W, A, Y, fz, tSeq, M=1000, Z=NA){
    require ("refund");
    require ("MASS");
    fpcaFit = fpca.sc (Y=W,var=TRUE, pve=0.99);
    lenT = dim(W)[2];
    tGrid = seq (0, 1, length=lenT);  
    Lam = fpcaFit$evalues/lenT;
    Phi = fpcaFit$efunctions*sqrt(lenT); 
    vig = fpcaFit$sigma2;
    
    ## Interpolate values of eigenfunctions at specified t-values
    interpPhi = function (tSeq){
        lnpc = dim(Phi)[2];
        nt = dim(Phi)[1];
        tTrain = seq (0, 1, length=nt);
        retPhi = matrix(NA, nrow=length(tSeq), ncol=lnpc);
        for (j in 1:lnpc){
            retPhi[,j] = approx (tTrain, Phi[,j], xout=tSeq, rule=2)$y;
        }
        return (retPhi);
    }

    interpMu = function (tSeq){
        nt = dim(Phi)[1];
        tTrain = seq (0, 1, length=nt);
        return (approx (tTrain, fpcaFit$mu, xout=tSeq, rule=2)$y);
    }

    ## Compute Gi 
    compG = function (WNew){
        ind = which (!is.na (WNew));
        Phii = Phi[ind,];
        return (Phii %*% diag(Lam) %*% t(Phii));
    }

   
    compGInterp = function (tSeq){
        lPhi = interpPhi (tSeq);
        return (lPhi %*% diag (Lam) %*% t(lPhi));
    }

    
    ## Estimate E(\xi_j \xi_l | W(T)) for a single W(T)
    prwsMnSngl = function (WNew){
        ind = which (!is.na (WNew));
        m = length (ind);
        Phii = Phi[ind,];
        PhiiLamRt = sqrt(Lam) * Phii;
        Im = diag (m);
        Gi = compG (WNew);
        Sig22 = vig*Im + Gi;
        Sig22Inv = solve (Sig22);
        npc = length (Lam);
        mnMat = matrix (0, nrow=npc, ncol=npc);
        for (j in 1:npc){
            for (k in 1:npc){
                if (j != k){
                    ## Compute components covariance for joint
                    ## distribution of \xi_j, \xi_k, W(T)
                    Sig11 = diag(c(Lam[j], Lam[k]));
                    Sig12 = rbind (Lam[j]*Phii[,j], Lam[k]*Phii[,k]);
                    mnjk = Sig12%*%Sig22Inv%*%(WNew[ind] -
                        as.numeric(fpcaFit$mu[ind]));
                    covjk = Sig11 - Sig12 %*% Sig22Inv %*% t(Sig12);
                    
                    ## Now compute conditional expectation
                    mnMat[j, k] = mnjk[1]*mnjk[2] + covjk[1,2];
                }
                else{
                    Sig12 = Lam[j]*Phii[,j];
                    mnjk = as.numeric (Sig12 %*% Sig22Inv) %*% (WNew[ind] -
                        as.numeric(fpcaFit$mu[ind]));
                    covjk = Lam[j] - Sig12 %*% Sig22Inv %*% Sig12;
                    mnMat[j,k] = mnjk^2 + covjk;
                }
            }
        }
        return (mnMat);
    }

    ## Compute E(\xi_i | W(T))
    getLoadingsSing = function (WNew){
        nLdings = numeric (length(Lam));
        ind = which(!is.na(WNew));
        m = length(ind);
        Phii = Phi[ind,]/sqrt(lenT);
        Lami = Lam*lenT;
        Im = diag (m);
        dgs = vig^(-1)*Im - vig^(-1)*Phii %*%
            solve (vig*diag(1/Lami) + t(Phii) %*% Phii) %*% t(Phii);
        nLdings = Lami*t(Phii) %*% dgs %*%
            (WNew[ind] - as.numeric(fpcaFit$mu[ind]))/sqrt(lenT);
        return (nLdings);
    }
    
    
    condMeanZ = function (WNew, tSeq){
        indW = which (!is.na(WNew));
        nW = length (indW);
        tW = tGrid[indW];
        nZ = length(tSeq);
        muZ = interpMu (tSeq);
        muW = interpMu (tW);
        phiKW = interpPhi (tW);
        phiKZ = interpPhi (tSeq);
        termI = matrix (0, nrow=nZ, ncol=nW);
        termII = matrix (0, nrow=nW, ncol=nW);
        npc = length (Lam);
        for (k in 1:npc){
            termI = termI + Lam[k]*phiKZ[,k] %*% t(phiKW[,k]);
            termII = termII + Lam[k]*phiKW[,k] %*% t(phiKW[,k]);
        }
        termII = ginv (termII + vig*diag(nW));
        retVal = muZ + termI %*% (termII %*% (WNew[indW] - muW));
        return (retVal);
    }

    condVarZ = function (WNew, tSeq){
        indW = which (!is.na(WNew));
        nW = length (indW);
        tW = tGrid[indW];
        nZ = length(tSeq);
        muZ = interpMu (tSeq);
        muW = interpMu (tW);
        phiKW = interpPhi (tW);
        phiKZ = interpPhi (tSeq);
        
        termI = matrix (0, nrow=nZ, ncol=nW);
        termII = matrix (0, nrow=nW, ncol=nW);
        termIII = numeric (length(tSeq));
        npc = length (Lam);
        for (k in 1:npc){
            termI = termI + Lam[k]*phiKZ[,k] %*% t(phiKW[,k]);
            termII = termII + Lam[k]*phiKW[,k] %*% t(phiKW[,k]);
            termIII = termIII + Lam[k]*phiKZ[,k] %*% t(phiKZ[,k]);
        }
        termII = ginv (termII + vig*diag(nW));
        retVal = termIII - termI %*% termII %*% t(termI);
        return (retVal);
    }
    

    ## Compute est of \mathbb{E}(f(Z(.) | W(T)))
    computeFixedFeature = function (WNew, tSeq){
        lPhi = interpPhi (tSeq);
        muZ = interpMu (tSeq);
        mn = condMeanZ ( WNew, tSeq);
        BCov = condVarZ (WNew, tSeq);
        ## Compute matrix square-root of BCov
        BEigen = eigen (BCov, symmetric=T);
        BSqrt = BEigen$vectors %*% diag(sqrt(pmax (BEigen$values,0))) %*%
            t (BEigen$vectors);
        ## Approximate conditional expectation
        efz = 0;
        nt = length (tSeq);
        for (b in 1:M){
            smplZ = mn + sigma*(BSqrt %*% rnorm (nt));
            efz = efz + fz (smplZ, tSeq);
        }
        return (efz/M);
    }
    
    ## Same as above but with multivariate WNew
    computeFixedFeatureMult = function (WNew, tSeq){
        nw = dim(WNew)[1];
        efz = numeric (nw);
        for (i in 1:nw){
            efz[i] = computeFixedFeature (WNew[i,], tSeq);
        }
        return (efz);
    }

    ## Compute ols estimator
    efz = computeFixedFeatureMult (W, tSeq);
    dMat = cbind(X, efz, A*X, A*efz);
    thetaHat = lm.fit (dMat, Y)$coefficients;
    p = dim (X)[2];
    alphaHat = thetaHat[1:p];
    nuHat = thetaHat[p+1];
    gammaHat = thetaHat[(p+2):(2*p+1)];
    etaHat = thetaHat[2*p+2];
    
    getTxts = function (XNew, WNew){
        efz = computeFixedFeatureMult (WNew, tSeq);
        pDMat = cbind (XNew, efz, XNew, efz);
        nDMat = cbind (XNew, efz, -XNew, -efz);
        return (sign((pDMat-nDMat)%*%thetaHat));
    }
    
    return (list("alphaHat"= alphaHat,
                 "nuHat"=nuHat,
                 "gammaHat"=gammaHat,
                 "etaHat"=etaHat,
                 "piHat"=getTxts,
                 "getFz"=computeFixedFeatureMult,
                 "Phi"=Phi,
                 "mu"=fpcaFit$mu));
}


## Same as above but does not make use of fnl predictors 
stdQLearn = function (X, W, A, Y){
    nObs = dim(W)[2];
    n = dim(W)[1];
    ## Last obs
    maxInd = cbind (1:n, unlist (lapply (apply (!is.na(W), 1, which), max)));
    dMatLast = cbind (X, W[maxInd], A*X, A*W[maxInd]);
    thetaLast = lm.fit (dMatLast, Y)$coefficients;
    getTxtsLast = function (XNew, WNew){
        nn = dim(WNew)[1];
        nMaxInd = cbind(1:nn,
            unlist (lapply (apply(!is.na(WNew), 1, which), max)));
        pDMat = cbind (XNew, WNew[nMaxInd], XNew, WNew[nMaxInd]);
        nDMat = cbind (XNew, WNew[nMaxInd], -XNew, -WNew[nMaxInd]);
        return (sign ((pDMat-nDMat)%*%thetaLast));
    }

    ## Mean obs
    WBar = apply(W, 1, mean, na.rm=T);
    dMatMean = cbind (X, WBar, A*X, A*WBar);
    thetaMean = lm.fit (dMatMean, Y)$coefficients;
    getTxtsMean = function (XNew, WNew){
        WNewMean = apply (WNew, 1, mean, na.rm=T);
        pDMat = cbind (XNew, WNewMean, XNew, WNewMean);
        nDMat = cbind (XNew, WNewMean, -XNew, -WNewMean);
        return (sign((pDMat - nDMat)%*%thetaMean));
    }
    return (list('piHatLast'=getTxtsLast,
                'piHatMean'=getTxtsMean));
}
