##
## @fn genSparseData.R
## @brief Generates sparsely observed longitudinal data
##        for use in txt selection.  The model is of the form:
##        Y = x^{\T}\alpha + \rho_{\beta}(Z) +
##              a(x^{\T}\gamma + \nu \rho_{\beta}(Z)) + \epsilon
##        where \rho_{\beta}(Z) = \int \beta(t)Z(t)dt is the linear
##        probe induced by \beta and epsilon is a normal error term.
## @author Anonymized
## @note Anonymized
##

##
## @fn genSparseData (n, obsPts, nBasis, bLoadings, p, alpha, gamma,
##                    nu, sigma, wnSigma, lam, kappa)
## @param n number of points to generate
## @param obsPts number of equaly spaced points on which *all*
##        observation times will fall, these must be in
##        the range (0,1)
## @param nBasis the number of basis functions 
## @param probe function defined on (0,1)
## @param p dimension of static predictors x
## @param alpha main effect coeffs for static predictors
## @param gamma interaction coeffs for static predictors
## @param nu scalar coeff for linear probe in main effect
## @param eta scalar coeff for linear probe in interaction
## @param sigma residual variance
## @param wnSigma white noise process variance
## @param lam nonnegative integer indexing poisson number of
##        observations per subject, deprecated
## @param kappaMin minimum number of pts to sample for each subj
## @param kappaMax maximum number of pts to sample for each subj
## @return list with the following key-->value pairs:
##         X --> matrix static predictors
##         W --> matrix of functional predictors 
##         A --> vector of txts
##         Y --> vector or responses
##
genSparseData = function (n, obsPts, nBasis, probe, p, alpha,
    gamma, nu, eta, sigma, wnSigma, lam, kappaMin, kappaMax){
    require ('fda');
    ## Generate non-fnl predictors
    X = cbind(1, matrix (rnorm(n*(p-1)), nrow=n));
    A = sign (rnorm(n));
    
    ## Generate fnl predictors
    if (max(obsPts) > 1 | min(obsPts) < 0){
        stop ('obsPts must belong to [0,1]');
    }
    bsb = create.bspline.basis (range=c(0,1), nbasis=nBasis);
    W = matrix (0, nrow=n, ncol=length(obsPts));
    lProbe = numeric (n);
    probeVals = probe (obsPts);
    betaStar = as.numeric (probeVals %*% eval.basis (obsPts, bsb)/
        length (probeVals));
    
    mu = rep(0, length(obsPts));
    wOptA = numeric (n);
    for (i in 1:n){
        randCoeff = rnorm (nBasis);
        z = mu + eval.basis (obsPts, bsb) %*% randCoeff;
        lProbe[i] = mean(z*probeVals);
        err = rnorm(length(obsPts))*wnSigma;
        W[i,] = z + err;
        nSamples = length(obsPts) - sample (kappaMin:kappaMax, size=1);
        exclInd = sample(1:length(obsPts), size=nSamples);
        inclInd = setdiff (1:length(obsPts), exclInd);
        phiT = eval.basis (obsPts[inclInd], bsb);
        
        GT = phiT %*% t(phiT);  # note, using var(\xi_k) = 1
        GTIInv = solve (GT + diag (length (inclInd)));
        xiMn = numeric (nBasis);
        for (l in 1:nBasis){
            xiMn[l] = phiT[,l] %*% GTIInv%*%W[i,inclInd];
        }
        wOptA[i] = sign (sum (X[i,]*gamma) + eta*sum (xiMn %*%betaStar));
        W[i, exclInd] = NA;  
    }
    Y = X %*% alpha + nu*lProbe + A*(X%*%gamma + eta*lProbe) + rnorm(n)*sigma;
    optA = sign (X%*%gamma + eta*lProbe);
    return (list(X=X, W=W, A=A, Y=Y, probeVals=lProbe, optA=optA,
                 wOptA=wOptA));
}


usage = function (){
    n = 100;
    nObsPts = 50;
    obsPts = seq (0, 1, length=nObsPts);
    nBasis = 4;
    probe = function (x){
        return (x-0.5);
    }
    p = 5;
    alpha = rep (1, p);
    gamma = rep (1, p);
    nu = 1;
    eta = 2;
    sigma=1.0;
    wnSigma = 0.5;
    lam = nObsPts*0.5;
    kappa = n;
    op = genSparseData (n, obsPts, nBasis, probe, p, alpha,
        gamma, nu, eta, sigma, wnSigma, lam, kappa);
    browser ();
}





